package com.jelurida.ardor.contracts;

import nxt.account.Account;
import nxt.addons.AbstractContract;
import nxt.addons.AbstractContractContext;
import nxt.addons.BlockContext;
import nxt.addons.ChainWrapper;
import nxt.addons.JA;
import nxt.addons.JO;
import nxt.addons.RequestContext;
import nxt.addons.TransactionContext;
import nxt.crypto.EncryptedData;
import nxt.http.callers.GetBlockchainStatusCall;
import nxt.http.callers.GetCoinExchangeTradesCall;
import nxt.http.callers.SendMessageCall;
import nxt.http.callers.SetAccountPropertyCall;
import nxt.util.Convert;

import java.math.BigDecimal;
import java.math.RoundingMode;

/**
 * Calculate the Ignis per Ardor exchange rate based on information from Bittrex and the Coin Exchange
 * Can be used to identify arbitrage opportunities
 */
public class IgnisArdorRates extends AbstractContract {

    /**
     * Write the result as account property
     */
    @Override
    public void processBlock(BlockContext context) {
        // Read contract configuration
        int frequency;
        if (getContractParams().isExist("frequency")) {
            frequency = getContractParams().getInt("frequency");
            if (frequency == 0) {
                return;
            }
        } else {
            JO blockchainStatus = GetBlockchainStatusCall.create().call();
            if (blockchainStatus.getBoolean("isTestnet")) {
                frequency = 3;
            } else {
                frequency = 60;
            }
        }
        // Only execute the contract on certain blocks
        if (context.getBlock().getHeight() % frequency != 0) {
            return;
        }

        // Get the data
        JO response = getTradeData(context);

        // Set account property with the data to be used by other contracts
        SetAccountPropertyCall setAccountPropertyCall = SetAccountPropertyCall.create(context.getChain("IGNIS").getId()).
                recipient(context.getConfig().getAccount()).property("IgnisPerArdorRates").value(response.toJSONString());
        context.createTransaction(setAccountPropertyCall);
    }

    /**
     * Send the result as a message to the sender
     */
    @Override
    public void processTransaction(TransactionContext context) {
        // Make sure the user paid 1 IGNIS or more for getting the data
        ChainWrapper ignisChain = context.getChain("IGNIS");
        if (context.notSameRecipient() || context.notSameChain(ignisChain.getId())) {
            return;
        }
        if (context.getAmountNQT() < ignisChain.getOneCoin()) {
            context.setErrorResponse(10001, "Oracle requires a payment of 1 IGNIS to operate");
            return;
        }

        // Get the data
        JO response = getTradeData(context);

        // Encrypt the message
        EncryptedData encryptedData = context.encryptTo(Account.getPublicKey(context.getSenderId()), Convert.toBytes(response.toJSONString(), true), context.getConfig().getSecretPhrase(), true);

        // Send a message back to the user who requested the information
        SendMessageCall sendMessageCall = SendMessageCall.create(context.getChainOfTransaction().getId()).recipient(context.getSenderId()).
                encryptedMessageData(encryptedData.getData()).
                encryptedMessageNonce(encryptedData.getNonce()).
                encryptedMessageIsPrunable(true);
        context.createTransaction(sendMessageCall);
    }

    /**
     * Send the result in response to API call
     */
    @Override
    public void processRequest(RequestContext context) {
        // Get the data and send it back to the invoking user
        // Requires admin password
        JO response = getTradeData(context);
        context.setResponse(response);
    }

    /**
     * Get the last trade data from Bittrex and from the coin exchange
     * @param context contract context
     * @return trade data
     */
    private JO getTradeData(AbstractContractContext context) {
        // Connect to Bittrex, the response format is listed below
        // {"success":true,"message":"","result":{"Bid":0.00003072,"Ask":0.00003098,"Last":0.00003090}}
        JO response = new JO();

        // Last IGNIS/BTC trade
        BittrexRateProvider bittrexRateProvider = new BittrexRateProvider();
        JO ignisTickerResponse = bittrexRateProvider.getRate(context, "IGNIS");
        if (ignisTickerResponse.get("errorCode") != null) {
            return response;
        }
        Double ignisLastTrade = ignisTickerResponse.getJo("result").numericToDouble("Last");

        // Last Ardor/BTC trade
        JO ardorTickerResponse = bittrexRateProvider.getRate(context, "ARDR");
        if (ardorTickerResponse.get("errorCode") != null) {
            return response;
        }
        Double ardorLastTrade = ardorTickerResponse.getJo("result").numericToDouble("Last");

        // Calculate IGNIS to Ardor rate
        long ignisNQTPerARDR = BigDecimal.valueOf(ardorLastTrade).
                multiply(BigDecimal.valueOf(context.getChain("IGNIS").getOneCoin())).
                divide(BigDecimal.valueOf(ignisLastTrade), RoundingMode.HALF_EVEN).
                longValue();
        response.put("BTRX", ignisNQTPerARDR);

        // Last IGNIS to Ardor trade from the coin exchange
        JO getCoinExchangeTradesResponse = GetCoinExchangeTradesCall.create(2).exchange(1).firstIndex(0).lastIndex(1).call();
        JA trades = getCoinExchangeTradesResponse.getArray("trades");
        if (trades.size() > 0) {
            JO trade = trades.get(0);
            response.put("CE", trade.get("priceNQTPerCoin"));
        }
        return response;
    }

}
